#!/usr/bin/env python
#coding:utf-8
"""
  Author:   --<v1ll4n>
  Purpose: iter mix
  Created: 03/09/17
"""

import unittest
import ast

forcode_tmp = """
for {valuebase}{index} in {iterbase}{index}:
    pass
try:
    {iterbase}{index}.reset() # {last_index}
except AttributeError:
    pass
"""

yieldcode_tmp = """
yield tuple([{content}])
"""

define_tmp = """
{iterbase}{index} = args[{index}]
"""

func_tmp = """
def dynamic_func(*args):
    pass
"""

mod_tmp = """
"""

mod = ast.parse(mod_tmp, '<string>', 'exec')


ITERBASE = 'arg_'
VALUEBASE = 'i_'

#
# compact components to a function
#----------------------------------------------------------------------
def _compact_dynamic_func(body):
    """"""
    mod = ast.parse(func_tmp, '<string>')
    funcnode = mod.body[0]
    funcnode.body.pop()
    for i in body:
        funcnode.body.append(i)
    
    return funcnode

#
# render the for code
#----------------------------------------------------------------------
def _render_for(iter_id_base, value_id_base, index, filled_expr=None):
    """"""
    iter_id_base = str(iter_id_base)
    value_id_base = str(value_id_base)
    index = int(index)
    lastindex = index if index - 1 < 0 else index - 1  
    
    _tmp = forcode_tmp.format(valuebase=value_id_base,
                              index=index,
                              iterbase=iter_id_base,
                              last_index=lastindex)
    
    _body = ast.parse(_tmp, '<string>').body
    _for = _body[0]
    
    if filled_expr:
        _for.body.pop()
        if isinstance(filled_expr, (list, tuple)):
            for i in filled_expr:
                _for.body.append(i)
        else:
            _for.body.append(filled_expr)
    return _body
    

#----------------------------------------------------------------------
def _render_yield(base, length):
    """"""
    base = str(base)
    length = int(length)
    ret = []
    for i in range(length):
        ret.append((base + '{}').format(i))
    yieldcode_new = yieldcode_tmp.format(content=','.join(ret))
    #print yieldcode_new
    return ast.parse(yieldcode_new, '<string>').body.pop()

#
# section2 yield
#----------------------------------------------------------------------
def _render_core_block(iter_num, iter_namebase=ITERBASE, value_namebase=VALUEBASE):
    """"""
    last = None
    for i in range(iter_num):
        i = iter_num - i - 1
        if last == None:
            last = _render_for(iter_namebase, value_namebase, i, _render_yield(value_namebase, iter_num))
        else:
            last = _render_for(iter_namebase, value_namebase, i, last)
    
    last = last[0]
    return last

#
# section1 define
#----------------------------------------------------------------------
def _define_iter(iterbase, length):
    """"""
    defnieblock = []
    for i in range(length):
        define_new = define_tmp.format(index=i, iterbase=iterbase)
        definebody = ast.parse(define_new, '<string>').body.pop()
        defnieblock.append(definebody)
    
    return defnieblock


#----------------------------------------------------------------------
def iter_mix(*args):
    """"""
    length = len(args)
    basebody = _define_iter(ITERBASE, length)
    basebody.append(_render_core_block(iter_num=length))
    funcmod = _compact_dynamic_func(basebody)
    mod_tmp = """
    """
    
    mod = ast.parse(mod_tmp, '<string>', 'exec')
    
    astt = funcmod
    if not isinstance(astt, (list, tuple)):
        mod.body.append(astt)
    else:
        for i in astt:
            mod.body.append(i)    
    
    exec(compile(mod, '<string>', 'exec'))
    return dynamic_func(*args)


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
#coding:utf-8
"""
  Author:   --<v11l4n>
  Purpose: Print Better Helper!
  Created: 2017/1/2
"""

import unittest
from colorama import init, Fore
from prettytable import PrettyTable

#----------------------------------------------------------------------
def print_bar(text=None, basic_char='=', length=70, color=None):
    """"""
    if text:
        text = str(text)
        textlen = len(text)
        length_one_side = (length - textlen) / 2
        bar_side = basic_char * int(length_one_side)
        bar = bar_side + text + bar_side
    else:
        bar = basic_char * length

    if color and hasattr(Fore, color.upper()):
        bar = getattr(Fore, color.upper()) + bar
    else:        
        bar = bar
    
    init(autoreset=True)
    print(bar)

#----------------------------------------------------------------------
def print_red(*args):
    """"""
    raw = str(args)
    init(autoreset=True)
    print((Fore.RED + raw))

#----------------------------------------------------------------------
def print_green(*args):
    """"""
    raw = str(args)
    init(autoreset=True)
    print((Fore.GREEN + raw))

#----------------------------------------------------------------------
def print_blue(*args):
    """"""
    raw = str(args)
    init(autoreset=True)
    print((Fore.CYAN + raw))
    
#----------------------------------------------------------------------
def print_yellow(*args):
    """"""
    raw = str(args)
    init(autoreset=True)
    print((Fore.YELLOW + raw))  

#----------------------------------------------------------------------
def print_cyan(*args):
    """"""
    raw = str(args)
    init(autoreset=True)
    print((Fore.CYAN + raw))  

#----------------------------------------------------------------------
def print_purpul(*args):
    """"""
    raw = str(args)
    init(autoreset=True)
    print((Fore.MAGENTA + raw))

#----------------------------------------------------------------------
def print_gray(*args):
    """"""
    raw = str(args)
    init(autoreset=True)
    print((Fore.LIGHTBLACK_EX + raw))
    

#----------------------------------------------------------------------
def print_column(head, column, color=''):
    """Print Column With Color.
    
    Params:
        head: :: A head for the column data.
        column: :list tuple: the column data(list or tuple)
        coler: :str: 
            lightblack_ex
            magenta
            cyan
            green
            blue
            yellow
            red
    
    Returns:
        return the strings of table 
    """
    assert isinstance(color, str)
    
    _prefix = ''
    if hasattr(Fore, color.upper()):
        _prefix = getattr(Fore, color.upper())
    
    table = PrettyTable()
    table.add_column(head, column)
    
    raw = table.get_string()
    init(autoreset=True)
    print(_prefix + raw)
    return table

#----------------------------------------------------------------------
def print_columns(heads, columns, color=''):
    """Print Column With Color.
    
    Params:
        head: :: A list of heads for the columns data.
        column: :list tuple: the column data(list or tuple)
        coler: :str: 
            lightblack_ex
            magenta
            cyan
            green
            blue
            yellow
            red
    
    Returns:
        return the strings of table 
    """
    assert isinstance(color, str)
    assert len(heads) == len(columns)
    
    _prefix = ''
    if hasattr(Fore, color.upper()):
        _prefix = getattr(Fore, color.upper())
    
    table = PrettyTable()
    for i in xrange(len(heads)):
        head = heads[i]
        column = columns[i]
        table.add_column(head, column)
    
    raw = table.get_string()
    init(autoreset=True)
    print(_prefix + raw)
    return table    
    
#----------------------------------------------------------------------
def print_rows(heads, rows, color=''):
    """Print Column With Color.
    
    Params:
        heads: :list: the row of heads
        rows: :list tuple: the rows data(list or tuple)
        coler: :str: 
            lightblack_ex
            magenta
            cyan
            green
            blue
            yellow
            red
    
    Returns:
        return the strings of table 
    """
    assert len(heads) == len(rows[0])
    assert isinstance(color, str)
    
    _prefix = ''
    if hasattr(Fore, color.upper()):
        _prefix = getattr(Fore, color.upper())
    
    #
    # split rows
    #
    table = PrettyTable(heads)
    for row in rows:
        table.add_row(row)
    
    
    raw = table.get_string()
    init(autoreset=True)
    print(_prefix + raw)
    return table        



if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
#coding:utf-8
"""
  Author:   --<v1ll4n>
  Purpose: Tester For Threadpoolex
  Created: 05/13/17
"""

import unittest
import queue
import time
from threadpoolex import ThreadPoolXLabor, ThreadPoolX, _LaborFactory

#----------------------------------------------------------------------
def test_task_error(arg1, arg2=4):
    """"""
    time.sleep(3)
    1/0
    return arg1, arg2

#----------------------------------------------------------------------
def test_task(arg1, arg2=4):
    """"""
    time.sleep(3)
    return arg1, arg2


########################################################################
class ThreadPoolExTester(unittest.TestCase):
    """"""

    #----------------------------------------------------------------------
    def test_labor(self):
        """"""
        def _labor_callback(result):
            raise ValueError()
            #_retqueue.put(result)

        def _labor_callback_exc(result):
            _retqueue.put(result)

        def _task_exc_handle(e):
            _retqueue.put((1,4))
            print e

        _retqueue = queue.Queue(1)
        s = ThreadPoolXLabor('test')
        s.add_task_exception_callback(_task_exc_handle)
        s.add_callback(_labor_callback, _labor_callback_exc)

        s.start()
        s.execute(test_task, var_args=(1,), keyword_args={})

        _resultTuple = _retqueue.get()
        self.assertTrue(_resultTuple[0] == 1)
        self.assertEqual(_resultTuple[1], 4)

        s.quit()
        time.sleep(1)


    #----------------------------------------------------------------------
    def test_pool(self):
        """"""
        quited = False
        def print_result(result):
            if result[0] == 5:
                print('prepared exiting')
                time.sleep(1)
                print('quit!')
                quited = True
            print result
        pool = ThreadPoolX()
        pool.add_callbacks(callback=print_result)
        for i in range(16):
            pool.feed(target=test_task, vargs=(i,))
        pool.start()

        def target():
            time.sleep(1)
            return 'h'

        def anothor_result(result):
            if result == 'h':
                print 'target with callback success!'

        pool.feed_with_callback(target, callback=anothor_result)
        time.sleep(6)
        pool.quit()

    #----------------------------------------------------------------------
    def test_laborfactory(self):
        """"""
        lf = _LaborFactory(debug=True, loop_interval=0.2)

        def _labor_callback(result):
            raise ValueError()
            #_retqueue.put(result)

        def _labor_callback_exc(result):
            _retqueue.put(result)

        def _task_exc_handle(e):
            _retqueue.put((1,4))
            print e

        _retqueue = queue.Queue(1)        
        lf.add_callbacks(_labor_callback, _labor_callback_exc)
        lf.add_exception_callback(_task_exc_handle)
        new_labor = lf.build_labor()

        self.assertIsInstance(new_labor, ThreadPoolXLabor)

        new_labor.start()
        new_labor.execute(test_task)
        new_labor.execute(test_task_error)
        new_labor.quit()


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
#coding:utf-8
"""
  Author:   --<v1ll4n>
  Purpose: Parse dict
  Created: 2016/12/16
"""

import pickle
import unittest
import os
import io
import types
from codecs import open
from pprint import pprint
from hashlib import md5

SESSION_TABLE_FILE = 'sessions_dat'
DEFAULT_SESSION_ID = 'default'

#----------------------------------------------------------------------
def get_buffer(shelvefile, key=None):
    """"""
    if os.path.exists(shelvefile):
        pass
    else:
        if key:
            set_buffer(shelvefile, key, value=0)
        else:
            set_buffer(shelvefile, key='default', value=0)

    with open(shelvefile, 'rb',) as fp:
        try:
            text = fp.read()
            pdict = pickle.loads(text)
            assert isinstance(pdict, dict)
            if key:
                result = pdict.get(key)
            else:
                result = pdict            
        except:
            if key:

                result = None
            else:
                result = {}



    return result


#----------------------------------------------------------------------
def set_buffer(shelvefile, key, value):
    """"""
    if os.path.exists(shelvefile):
        pdict = get_buffer(shelvefile)
    else:
        pdict = {}

    pdict[key] = value
    text = pickle.dumps(pdict)

    with open(shelvefile, 'wb') as fp:
        fp.write(text)





########################################################################
class DictParserFromIter(object):
    """"""

    #----------------------------------------------------------------------
    def __init__(self, list_or_content, 
                 session_id=SESSION_TABLE_FILE,
                 do_continue=False,
                 session_data_file=SESSION_TABLE_FILE):
        """Constructor

        Params:
            list_or_content: :str: the target dict you want to use
            session_id: :str: the id you want to use to identify your session
            do_continue: :bool: continue or not
            session_data_file: :str: the file you want to save session info
            """
        
        if isinstance(list_or_content, (set, list, tuple)):
            _list = list(list_or_content)
            self._content = '\n'.join(_list)
        elif isinstance(list_or_content, types.StringTypes):
            self._content = list_or_content
        
        #filename = 'default'
        #abspathfile = os.path.abspath(filename)
        #if os.path.exists(abspathfile):
            #self._filename = filename
        #else:
            #raise Exception('[!] No Such Dict File')

        self._session_data_file = session_data_file
        self._session_id = md5(self._content).hexdigest()

        try:
            self._session_progress_table = get_buffer(self._session_data_file)
        except:
            os.remove(os.path.abspath(self._session_data_file))
            self._session_progress_table = get_buffer(self._session_data_file)

        self._do_continue = do_continue

        if hasattr(self, '_dict_file_p'):
            pass
        else:
            self.__enter__()

    #----------------------------------------------------------------------
    def __enter__(self):
        """"""
        self._dict_file_p = io.StringIO()
        self._dict_file_p.write(unicode(self._content))
        self._dict_file_p.seek(0)

        # continue last task
        if self._do_continue:
            if self._session_id in self._session_progress_table:
                try:
                    pos = get_buffer(self._session_data_file, self._session_id)
                except ValueError:
                    pos = 0
                self._dict_file_p.seek(pos)      

        return self

    #----------------------------------------------------------------------
    def __exit__(self):
        """"""
        self._dict_file_p.close()
        return True




    #----------------------------------------------------------------------
    def __iter__(self):
        """"""
        return self


    #----------------------------------------------------------------------
    def __next__(self):
        """"""
        return self._next()

    #----------------------------------------------------------------------
    def next(self):
        """"""
        return self._next()


    #----------------------------------------------------------------------
    def _next(self):
        """"""
        self._current_pos = 0
        self._current_pos = self._dict_file_p.tell()
        while True:
            ret = self._dict_file_p.readline()
            self.save()
            if ret.strip() == '':
                #_empty_lines_count = _empty_lines_count + 1
                if self._dict_file_p.tell() == self._current_pos:
                    raise StopIteration()
                else:
                    continue
            else:
                return ret.strip()

    #----------------------------------------------------------------------
    def save(self):
        """"""
        if self._current_pos - self._dict_file_p.tell() >= 1024:
            self._save()        
        #print self._session_progress_table


    #----------------------------------------------------------------------
    def _save(self):
        """Save the progress"""
        #self._session_progress_table[self._session_id] = self._dict_file_p.tell()

        set_buffer(self._session_data_file, self._session_id, self._dict_file_p.tell())

    #----------------------------------------------------------------------
    def force_save(self):
        """Force save"""
        self._save()


    #----------------------------------------------------------------------
    def __del__(self):
        """Close the opened resource"""
        try:
            self.close()
            #self._session_progress_table.close()
        except AttributeError:
            pass

    #----------------------------------------------------------------------
    def close(self):
        """"""
        self.force_save()

        if self._dict_file_p.closed:
            pass
        else:
            self._dict_file_p.close()


    #----------------------------------------------------------------------
    def get_next_collection(self, num=200):
        """Returns a tuple for the next [num] paylaod(lines)"""
        ret = []
        for i in range(num):
            try:
                ret.append(self._next())
            except StopIteration:
                pass
        self._save()

        return tuple(ret)

    #----------------------------------------------------------------------
    def get_current_pos(self):
        """The current pos"""
        return self._dict_file_p.tell()

    #----------------------------------------------------------------------
    def get_total_size(self):
        """total size(progress)"""
        return os.path.getsize(self._filename)

    #----------------------------------------------------------------------
    def get_fp(self):
        """Get the fp(Dangerous)"""
        return self._dict_file_p

    #----------------------------------------------------------------------
    def reset(self):
        """"""
        self._dict_file_p.seek(0)

########################################################################
class TESTer(unittest.case.TestCase):
    """"""

    #----------------------------------------------------------------------
    def test_dpfromlistorcontent(self):
        """"""
        content = ['1123123123123', 'sssssssssssssss', 'ssssssssss',
                   'sssssssssssssss', 'ssssssssss']
        
        dp = DictParserFromIter(content, do_continue=False)
        self.assertTrue(dp.next() == '1123123123123')
        dp.reset()
        self.assertTrue(dp.next() == '1123123123123')
        dp.save()
        
        dp = DictParserFromIter(content, do_continue=True)
        self.assertTrue(dp.next() == u'sssssssssssssss')
        



if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
#coding:utf-8
"""
  Author:   --<v1ll4n>
  Purpose: G3ar entry
  Created: 2016/12/20
"""

import unittest

from .threadutils.thread_pool import Pool as ThreadPool
from .threadutils.threadpoolex import ThreadPoolX
from .threadutils.contractor import Contractor
from .decologger.decologger import Decologger as DecoLogger
from .dict_parser.dict_parser import DictParser
from .dict_parser.dict_parser_from_iter import DictParserFromIter
from .taskbulter.task_bulter import TaskBulter
from .dict_parser.dict_parser_mixer import DictParserMixer
#from .utils import ip_calc_utils
from .utils.pyping import pyping
from .utils.iter_utils import iter_mix

#----------------------------------------------------------------------
def ping(host, timeout=2, count=4):
    """"""
    return pyping(host, timeout, count)

#----------------------------------------------------------------------
def dict_parser_mixer_func(*args):
    """"""
    for i in args:
        assert isinstance(i, DictParser)
    
    return iter_mix(*args)
    


#!/usr/bin/env python
#coding:utf-8
"""
  Author:   --<v1ll4n>
  Purpose: threads utils
  Created: 03/07/17
"""

import unittest
import threading

#----------------------------------------------------------------------
def start_new_thread(target, name=None, args=tuple(), 
                     kwargs={}, daemon=True):
    """"""
    ret = threading.Thread(target=target, args=args, 
                           kwargs=kwargs, name=name)
    ret.daemon = daemon
    ret.start()
    return ret
    
    

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
#coding:utf-8
"""
  Author:   --<v1ll4n>
  Purpose: ThreadPool From Twisted and Add Common Resource
  Created: 05/13/17
"""

from __future__ import unicode_literals

import uuid
import threading
try:
    import queue
except:
    import Queue as queue
import time
import random

class LaborQuit(Exception):
    pass


########################################################################
class ThreadPoolXLabor(threading.Thread):
    """"""

    #----------------------------------------------------------------------
    def __init__(self, name, debug=True, loop_interval=0.2):
        """Constructor"""
        threading.Thread.__init__(self, name=name)

        #
        # key flag
        #
        self.working = False
        self._debug = True
        self._waiting_quit = False

        #
        # cache flag
        #
        self._cb_cached = False
        self._ecb_cached = False

        #
        # fields
        #
        self.loop_interval = loop_interval

        #
        # private attributes
        #
        self._task_buffer = queue.Queue(1)
        self._callback_chains = []
        self._exception_handle_callback_chains = []

        self._cb_c_buffer = self._callback_chains
        self._ehcb_c_buffer = self._exception_handle_callback_chains

    #----------------------------------------------------------------------
    def run(self):
        """"""
        try:
            self._run()
        except LaborQuit as e:
            pass

    #----------------------------------------------------------------------
    def _run(self):
        """"""
        self.working = True
        while self.working:
            #
            # if quit, before get task
            #
            if self._waiting_quit:
                self._quit()
            #
            # get task from task_buffer
            # 
            _task = self._get_current_task()

            #
            # process _task
            #
            if _task:
                _callabled = _task[0]
                _vargs = _task[1]
                _kwargs = _task[2]

                #
                # define exception
                #
                exc = None
                try:
                    _result = _callabled(*_vargs, **_kwargs)
                    self._handle_result(_result)

                    #
                    # recovery callback chains
                    #
                    if self._cb_cached:
                        self._recovery_cb_chain()
                except Exception as e:
                    self._handle_exception(e)

                    #
                    # recovery error_callback 
                    #
                    if self._ecb_cached:
                        self._recovery_ecb_chain()

            else:
                if self._waiting_quit:
                    self._quit()
                time.sleep(self.loop_interval)

    #----------------------------------------------------------------------
    @property
    def busy(self):
        """"""
        if self._task_buffer.qsize() == 0:
            return False
        else:
            return True

    #----------------------------------------------------------------------
    def _handle_exception(self, exception_obj):
        """"""
        e = exception_obj
        #
        # exception callback
        #
        if self._exception_handle_callback_chains == []:
            if self._debug:
                raise e
        else:
            for i in self._exception_handle_callback_chains:
                e = i(e)         

    #----------------------------------------------------------------------
    def _handle_result(self, result):
        """"""
        _result = result

        #
        # process result callback
        #
        for i in self._callback_chains:

            _cb = i[0]
            _exc_cb = i[1]

            try:
                _result = _cb(_result)
            except Exception as e:
                #
                # if some exception happend
                #
                if _exc_cb:
                    _result = _exc_cb(_result)
                else:
                    pass        


    #----------------------------------------------------------------------
    def _get_current_task(self):
        """"""
        try:
            _ret = self._task_buffer.get_nowait()
        except:
            _ret = None

        return _ret

    #----------------------------------------------------------------------
    def add_callback(self, callback, exception_callback=None):
        """result callback function"""
        assert callable(callback), 'callback cannot be called'

        self._callback_chains.append((callback, exception_callback))

    #----------------------------------------------------------------------
    def add_task_exception_callback(self, callback):
        """"""
        assert callable(callback), 'exception callback'

        self._exception_handle_callback_chains.append(callback)


    #----------------------------------------------------------------------
    def _execute(self, target, var_args=tuple(), keyword_args={}):
        """"""
        assert callable(target), 'target function cannot be called'
        assert not self._waiting_quit, 'waiting for quiting, cannot execute task'

        self._task_buffer.put(tuple([target, var_args, keyword_args]))

    #----------------------------------------------------------------------
    def execute(self, target, var_args=tuple(), keyword_args={}):
        """"""
        if self._cb_cached:
            self._recovery_cb_chain()

        if self._ecb_cached:
            self._recovery_ecb_chain()

        self._execute(target, var_args, keyword_args)

    #----------------------------------------------------------------------
    def execute_with_callback(self, target, var_args=tuple(), keyword_args={},
                              callback=None, callback_exc=None, error_callback=None,):
        """"""
        assert callable(target), 'target function cannot be called'
        assert not self._waiting_quit, 'waiting for quiting, cannot execute task'

        #
        # process callback
        # 
        if callback:
            self._cache_cb_chain()
            self.add_callback(callback, callback_exc)

        if error_callback:
            self._cache_ecb_chain()
            self.add_task_exception_callback(error_callback)

        #
        # execute called
        #
        self._execute(target, var_args, keyword_args)


    #----------------------------------------------------------------------
    def execute_with_callback_chains(self, target, var_args=tuple(), keyword_args={},
                                     callback_chain=None, error_callback_chains=None):
        """"""
        assert callable(target), 'target function cannot be called'
        assert not self._waiting_quit, 'waiting for quiting, cannot execute task'
        assert callback_chain == None or isinstance(callback_chain, list), \
               'callback chain has to be None or list'
        assert error_callback_chains == None or isinstance(error_callback_chains, list), \
               'callback chain has to be None or list'

        #
        # proccess callback
        #
        if callback_chain:
            self._cache_cb_chain()
            self._callback_chains = callback_chain

        if error_callback_chains:
            self._cache_ecb_chain()
            self._exception_handle_callback_chains = error_callback_chains

        self._execute(target, var_args, keyword_args)

    #----------------------------------------------------------------------
    def _cache_cb_chain(self):
        """"""
        #
        # cache callback
        #
        self._cb_c_buffer = self._callback_chains

        #
        # change flag
        #
        self._cb_cached = True

        #
        # reset chains
        #
        self._callback_chains = []

    #----------------------------------------------------------------------
    def _recovery_cb_chain(self):
        """"""
        self._cb_cached = False
        self._callback_chains = self._cache_cb_chain


    #----------------------------------------------------------------------
    def _recovery_ecb_chain(self):
        """"""
        self._ecb_cached = False
        self._exception_handle_callback_chains = self._cache_ecb_chain     

    #----------------------------------------------------------------------
    def _cache_ecb_chain(self):
        """The same as self._cache_cb_chain"""
        self._ehcb_c_buffer = self._exception_handle_callback_chains
        self._ecb_cached = True
        self._exception_handle_callback_chains = []


    #----------------------------------------------------------------------
    def quit(self):
        """"""
        self._waiting_quit = True

    #----------------------------------------------------------------------
    def _quit(self):
        """"""
        raise LaborQuit('normal exit labor')


########################################################################
class _LaborFactory(object):
    """"""
    count = 0
    Labor = ThreadPoolXLabor

    #----------------------------------------------------------------------
    def __init__(self, debug, loop_interval):
        """"""

        self._loop_interval = loop_interval
        self._debug = debug

        self.callback_chains = []
        self.exception_callback_chains = []

    #----------------------------------------------------------------------
    def add_callbacks(self, callback, callback_exc=None):
        """"""
        assert callable(callback), 'callback not callable'
        assert callback_exc == None or callable(callback_exc), 'result exception callback not callable.'

        self.callback_chains.append(tuple([callback, callback_exc]))

    #----------------------------------------------------------------------
    def add_callback_chain(self, callback_chain):
        """"""
        assert isinstance(callback_chain, list), 'callback chain must be list'
        self.callback_chains = callback_chain

    #----------------------------------------------------------------------
    def add_exception_callback(self, callback):
        """"""
        assert callable(callback), 'exception callback cannot be called'

        self.exception_callback_chains.append(callback)

    #----------------------------------------------------------------------
    def add_exception_callback_chain(self, callback_chain):
        """"""
        assert isinstance(callback_chain, list), 'callback chain must be list'

        self.exception_callback_chains = callback_chain

    #----------------------------------------------------------------------
    def build_labor(self):
        """"""

        _lb = self.Labor(name=self.gen_labor_name(), 
                         debug=self._debug, 
                         loop_interval=self._loop_interval)

        #
        # add result callback
        #
        for i in self.callback_chains:
            _lb.add_callback(*i)

        #
        # add exception callback
        #
        for i in self.exception_callback_chains:
            _lb.add_task_exception_callback(i)

        return _lb

    #----------------------------------------------------------------------
    def gen_labor_name(self):
        """"""
        return 'labor-{index}_starttime:{ts}'.format(index=self.count, 
                                                     ts=int(time.time()))






########################################################################
class _ThreadTeam(object):
    """"""

    #----------------------------------------------------------------------
    def __init__(self, factory):
        """Constructor"""
        assert isinstance(factory, _LaborFactory)

        self._factory = factory

        self._labors = []

        #
        # selected labor lock
        #
        self._selected_labor_lock = threading.Lock()
        self._selected_labor = []

    #----------------------------------------------------------------------
    @property    
    def labors(self):
        """"""
        return self._labors

    #----------------------------------------------------------------------
    def add(self):
        """"""
        lb = self._factory.build_labor()
        #
        # start labor
        #
        lb.daemon = True
        lb.start()
        self.labors.append(lb)
        return lb

    #----------------------------------------------------------------------
    def shrink(self):
        """"""
        #
        # shrink
        #
        map(self._exit_idle_labor, self.labors)

    #----------------------------------------------------------------------
    def _exit_idle_labor(self, labor):
        """"""
        assert isinstance(labor, ThreadPoolXLabor)

        if labor.is_alive():
            if labor.busy:
                pass
            else:
                self.labors.remove(labor)
                labor.quit()
        else:
            self.labors.remove(labor)

    #----------------------------------------------------------------------
    @property
    def size(self):
        """"""
        return len(filter(lambda x: x.is_alive(), self.labors))

    #----------------------------------------------------------------------
    def quitall(self):
        """"""
        #
        # shrink first
        #
        self.shrink()

        #
        # quit all labor
        #
        for i in self.labors:
            i.quit()
            i.join()
            assert not i.is_alive()

    #----------------------------------------------------------------------
    def select(self):
        """"""
        _idle_labors = filter(lambda x: not x.busy, self.labors)

        if _idle_labors == []:
            return None
        else:
            labor = random.choice(_idle_labors)
            while labor in self._selected_labor:
                labor = random.choice(_idle_labors)

            self._selected_labor_lock.acquire()
            if labor in self._selected_labor:
                pass
            else:
                self._selected_labor.append(labor)
            self._selected_labor_lock.release()
            return labor

    #----------------------------------------------------------------------
    def release_labor(self, labor):
        """"""
        #
        # sync selected labor
        #
        self._selected_labor_lock.acquire()
        self._selected_labor.remove(labor)
        self._selected_labor_lock.release()



########################################################################
class ThreadPoolX(object):
    """"""

    min_size = 5
    max_size = 20

    joined = False
    started = False

    workers = 0
    name = 'default'

    #----------------------------------------------------------------------
    def __init__(self, min_threads=5, max_threads=20, 
                 name='default', debug=True, loop_interval=0.2,
                 adjuest_interval=3, diviation_ms=100):
        """Create threadpool object

        @param min_threads: min size of thread pool
        @type min_threads: L{int}

        @param max_threads: max size of thread pool
        @type max_threads: L{int}

        @param name: The name to give this threadpool; visible in log msg.
        @type name: native L{str}"""

        assert min_threads >= 0, 'minimum is negative'
        assert min_threads <= max_threads, 'maximum is less than minimum'

        self.min_size = min_threads
        self.max_size = max_threads

        self.name = name
        self.diviation_ms = diviation_ms
        #self.threads = []

        #
        # init factory
        #
        self._factory = _LaborFactory(debug, loop_interval)

        #
        # init _team
        #
        self._team = _ThreadTeam(self._factory)

        #
        # private entity
        #
        self._task_queue = queue.Queue()

        #
        # control adjust pool size interval
        #
        self._lp_itrvl = loop_interval
        self._adj_itrvl = adjuest_interval

        #
        # dispatcher
        #
        def _dispatcher_factory():
            _ret =  threading.Thread(name='ThreadPoolX:{name} Dispatcher'\
                                     .format(name=self.name), 
                                     target=self._start)

            _ret.daemon = True
            return _ret

        self._dispatcher = _dispatcher_factory()

    #
    # set callback
    #
    #----------------------------------------------------------------------
    def add_callbacks(self, callback, callback_exc=None):
        """"""
        self._factory.add_callbacks(callback, callback_exc)

    #----------------------------------------------------------------------
    def add_callback_chain(self, callback_chain):
        """"""
        self._factory.add_callback_chain(callback_chain)

    #----------------------------------------------------------------------
    def add_exception_callback(self, callback):
        """"""
        self._factory.add_exception_callback(callback)

    #----------------------------------------------------------------------
    def add_exception_callback_chain(self, callback_chain):
        """"""
        self._factory.add_exception_callback_chain(callback_chain)

    #----------------------------------------------------------------------
    def _start(self):
        """"""
        self.started = True
        self.adjust_pool_size()
        while self.started:
            if int(int(time.time()*1000000) % (self._adj_itrvl*1000000)) <= self.diviation_ms:
                #print 'adpool'

                self.adjust_pool_size()
            #print self.dumped_status()
            #
            # consume task
            #
            self.consume()
            #

    #----------------------------------------------------------------------
    def start(self):
        """"""
        self._dispatcher.start()

    #----------------------------------------------------------------------
    def join(self):
        """"""
        self._dispatcher.join()

    #----------------------------------------------------------------------
    def quit(self):
        """"""
        self.started = False
        self.join()
        assert not self._dispatcher.is_alive()
        print('Existed ThreadPool MainLoop(Dispatcher!)')
        self._team.quitall()
        print('Quit All Labor!')


    #----------------------------------------------------------------------
    def dumped_status(self):
        """"""
        state = {}
        state['name'] = self.name
        state['current_size'] = self._team.size
        state['idle_labor_size'] = len(filter(lambda x: x.is_alive() and not x.busy, self._team.labors))

        return state


    #----------------------------------------------------------------------
    def consume(self):
        """"""
        #
        # check team and select a idle labor
        #
        _labor = self._team.select()

        if _labor:
            pass
        else:
            if self._team.size < self.max_size:
                self._team.add() 
                _labor = self._team.select()
            else:
                _labor = None

        #
        # labor existed, execute task
        #
        if _labor:
            try:
                _task = self._task_queue.get_nowait()
            except queue.Empty:
                _task = None

            if _task:
                _labor.execute(*_task)

            self._team.release_labor(_labor)
        else:
            pass

    #----------------------------------------------------------------------
    def adjust_pool_size(self):
        """"""

        while self._team.size < self.min_size:
            self._team.add()

        if self._team.size > self.max_size:
            self._team.shrink()

    #----------------------------------------------------------------------
    def feed(self, target, vargs=tuple(), kwargs={}):
        """"""
        assert callable(target), 'target function cannot be executed'

        self._task_queue.put(tuple([target, vargs, kwargs]))

    #----------------------------------------------------------------------
    def feed_with_callback(self, target, vargs=tuple(), kwargs={}, 
                           callback=None, callback_exc=None, error_callback=None):
        """"""
        assert callable(target), 'target function cannot be executed'

        _temp_labor = self._team.add()
        _temp_labor.execute_with_callback(target, vargs, kwargs, callback,
                                          callback_exc, error_callback)

    #----------------------------------------------------------------------
    def feed_with_callback_chain(self, target, vargs=tuple(), kwargs={},
                                 callback_chain=None, error_callback_chain=None):
        """"""
        assert callable(target), 'target function cannot be executed'

        _temp_labor = self._team.add()
        _temp_labor.execute_with_callback_chains(target, vargs, kwargs,
                                                 callback_chain=callback_chain, 
                                                 error_callback_chains=error_callback_chain)

